import tempfile
from dataclasses import asdict, dataclass
from pathlib import Path

import scann
from safecheck import Float, Float32, NumpyArray, UInt32, typecheck
from scann.scann_ops.py import scann_ops_pybind_backcompat
from typing_extensions import Any, Literal, NotRequired, TypedDict, get_type_hints

from ._base import NearestNeighbors


@dataclass
class ScannTreeConfig:
    num_leaves: int = 1000
    num_leaves_to_search: int = 10
    training_sample_size: int = 100000
    min_partition_size: int = 50
    training_iterations: int = 10
    spherical: bool = False
    quantize_centroids: bool = False
    random_init: bool = True


@dataclass
class ScannHashingConfig:
    dimensions_per_block: int = 2
    anisotropic_quantization_threshold: float = 0.2
    training_sample_size: int = 100000
    min_cluster_size: int = 100
    hash_type: str = "lut16"
    training_iterations: int = 10


@dataclass
class ScannBruteForceConfig:
    quantize: bool = False


@dataclass
class ScannReorderConfig:
    reordering_num_neighbors: int = 100
    quantize: bool = False


@dataclass
class ScannSearchConfig:
    pre_reorder_num_neighbors: int | None = None
    leaves_to_search: int | None = None


def typedict(data_class: type) -> type[TypedDict]:  # type: ignore[reportGeneralTypeIssues]
    """Convert the dataclass to a typed dictionary.

    All dictionary items are set to ``NotRequired`` because we have default initializers
    for all config items where ``typedict`` is used.

    :param data_class: Configuration data class with default values.
    :return: TypedDict representation of the dataclass.
    """
    field_types: dict[str, type] = {
        k: NotRequired[t] for k, t in get_type_hints(data_class).items()  # type: ignore[reportGeneralTypeIssues]
    }
    return TypedDict(f"{data_class.__name__}Dict", field_types)  # type: ignore[reportGeneralTypeIssues]


ScannTreeConfigDict = typedict(ScannTreeConfig)
ScannBruteForceConfigDict = typedict(ScannBruteForceConfig)
ScannHashingConfigDict = typedict(ScannHashingConfig)
ScannReorderConfigDict = typedict(ScannReorderConfig)
ScannSearchConfigDict = typedict(ScannSearchConfig)


class ScannNeighbors(NearestNeighbors):
    """ScaNN performs search in three phases as described in the following.

    1. Partitioning (optional): ScaNN partitions the dataset during training time, and at query time selects the top
    partitions to pass onto the scoring stage.
    2. Scoring: ScaNN computes the distances from the query to all datapoints in the dataset
    (if partitioning isn't enabled) or all datapoints in a partition to search (if partitioning is enabled).
    These distances aren't necessarily exact, but brute-force scoring can be enabled.
    3. Rescoring (optional): ScaNN takes the best k' distances from scoring and re-computes these distances more
    accurately. From these k' re-computed distances the top k are selected.

    Source: https://github.com/google-research/google-research/blob/master/scann/docs/algorithms.md
    """

    available_metrics = Literal["dot_product", "squared_l2"]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "squared_l2",
        n_neighbors: int = 1,
        n_training_threads: int = 0,
        use_tree: bool = True,
        use_reorder: bool = False,
        use_bruteforce: bool = False,
        search_parallel: bool = False,
        tree_config: ScannTreeConfig | ScannTreeConfigDict | None = None,  # type: ignore[reportGeneralTypeIssues]
        bruteforce_config: ScannBruteForceConfig | ScannBruteForceConfigDict | None = None,  # type: ignore[reportGeneralTypeIssues]
        hashing_config: ScannHashingConfig | ScannHashingConfigDict | None = None,  # type: ignore[reportGeneralTypeIssues]
        reorder_config: ScannReorderConfig | ScannReorderConfigDict | None = None,  # type: ignore[reportGeneralTypeIssues]
        search_config: ScannSearchConfig | ScannSearchConfigDict | None = None,  # type: ignore[reportGeneralTypeIssues]
    ) -> None:
        super().__init__()
        # safely initialize the mutable config parameters
        to_config = lambda c, v: c() if v is None else c(**v) if isinstance(v, dict) else v
        self.parameters.tree_config = to_config(ScannTreeConfig, tree_config)
        self.parameters.bruteforce_config = to_config(ScannBruteForceConfig, bruteforce_config)
        self.parameters.hashing_config = to_config(ScannHashingConfig, hashing_config)
        self.parameters.reorder_config = to_config(ScannReorderConfig, reorder_config)
        self.parameters.search_config = to_config(ScannSearchConfig, search_config)

        # to be defined in ``fit``
        self._index = None

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "ScannNeighbors":
        parameters = self.parameters
        score_config = parameters.bruteforce_config if parameters.use_bruteforce else parameters.hashing_config
        score_method = "score_brute_force" if parameters.use_bruteforce else "score_ah"

        # set up the builder
        searcher = scann.scann_ops_pybind.builder(
            db=data,
            num_neighbors=parameters.n_neighbors,
            distance_measure=parameters.metric,
        )

        # set number of threads
        searcher.set_n_training_threads(self.parameters.n_training_threads)

        # provide the tree config
        if self.parameters.use_tree:
            searcher.tree(**asdict(parameters.tree_config))

        # provide the score config
        getattr(searcher, score_method)(**asdict(score_config))

        # provide the reorder config
        if self.parameters.use_reorder:
            searcher.reorder(**asdict(parameters.reorder_config))

        self._index = searcher.build()
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[UInt32[NumpyArray, "{n_neighbors}"], Float32[NumpyArray, "{n_neighbors}"]]:
        idx, dist = self._index.search(
            point,
            final_num_neighbors=n_neighbors,
            **asdict(self.parameters.search_config),
        )
        return idx, dist

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[UInt32[NumpyArray, "m {n_neighbors}"], Float32[NumpyArray, "m {n_neighbors}"]]:
        search_method = "search_batched_parallel" if self.parameters.search_parallel else "search_batched"
        idx, dist = getattr(self._index, search_method)(
            points,
            final_num_neighbors=n_neighbors,
            **asdict(self.parameters.search_config),
        )
        return idx, dist

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Save the bytes to files and read the index from the temporary files."""
        super().__setstate__(state)
        if self._index is not None:
            index_data = state.pop("_index")
            with tempfile.TemporaryDirectory() as tmp_dir:
                tmp_path = Path(tmp_dir)
                for name, data in index_data.items():
                    (tmp_path / name).write_bytes(data)

                # override the scann_assets.pbtxt file in the current directory, otherwise the paths to the
                # assets include the old temporary directory prefixes
                scann_ops_pybind_backcompat.populate_and_save_assets_proto(tmp_dir)
                self._index = scann.scann_ops_pybind.load_searcher(tmp_dir)

    def __getstate__(self) -> dict[str, Any]:
        """Save the index to temporary directory and store the resulting bytes in a dictionary to re-build the index."""
        state = super().__getstate__()
        if self._index is not None:
            with tempfile.TemporaryDirectory() as tmp_dir:
                self._index.serialize(tmp_dir)
                tmp_path = Path(tmp_dir)
                files = tmp_path.glob("*")
                data = {file.name: file.read_bytes() for file in files}
                state["_index"] = data

        return state



import numpy as np
from safecheck import Float, Int64, NumpyArray, typecheck
from typing_extensions import Literal

from ._base import NearestNeighbors


def cdist_minkowski(
    data: Float[NumpyArray, "n d"],
    query: Float[NumpyArray, "m d"],
    p: int | float = 2,
) -> Float[NumpyArray, "m n"]:
    """Reference implementation of Euclidean distance with NumPy.

    :param data: Matrix of input data points.
    :param query: Matrix of query data points.
    :param p: The power parameter for the Minkowski distance metric.
    :return: Matrix of distances between query and data.
    """
    diff = query[:, np.newaxis, :] - data[np.newaxis, :, :]
    return np.linalg.norm(diff, ord=p, axis=-1, keepdims=False)


def cdist_euclidean_mm(
    data: Float[NumpyArray, "n d"],
    query: Float[NumpyArray, "m d"],
) -> Float[NumpyArray, "m n"]:
    """Probably the fastest possible implementation of Euclidean distance with NumPy.

    :param data: Matrix of input data points.
    :param query: Matrix of query data points.
    :return: Matrix of distances between query and data.
    """
    # Calculate the squared differences raised to the power of p
    squared_diff_p = (
        np.einsum("ij,ij->i", data, data)[np.newaxis, :]
        - 2.0 * np.dot(query, data.T)
        + np.einsum("ij,ij->i", query, query)[:, np.newaxis]
    )

    # Ensure that negative values close to zero are treated as zeros due to potential numerical errors
    np.maximum(squared_diff_p, 0.0, out=squared_diff_p)
    return np.sqrt(squared_diff_p)


def min_k(arr: Float[NumpyArray, "m n"], k: int) -> tuple[Int64[NumpyArray, "m {k}"], Float[NumpyArray, "m {k}"]]:
    """Similar to torch.topk, but only implemented for matrices and returning the smallest (min) items.

    :param arr: Matrix of data points (distances).
    :param k: Number of smallest elements to determine.
    """
    indices = np.argpartition(arr, kth=range(k), axis=1)[:, :k]
    values = np.take_along_axis(arr, indices, axis=1)
    return indices, values


class NumpyNeighbors(NearestNeighbors):
    """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""

    available_metrics = Literal["minkowski"]
    compute_modes = Literal["use_mm_for_euclid_dist", "donot_use_mm_for_euclid_dist"]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "minkowski",
        p: int = 2,
        compute_mode: compute_modes = "use_mm_for_euclid_dist",
    ) -> None:
        super().__init__()
        # to be defined in ``fit``
        self._data: Float[NumpyArray, "n d"] | None = None

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "NumpyNeighbors":
        self._data = data
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float[NumpyArray, "{n_neighbors}"]]:
        """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], Float[NumpyArray, "m {n_neighbors}"]]:
        if self.parameters.p == 2 and self.parameters.compute_mode == "use_mm_for_euclid_dist":
            distance = cdist_euclidean_mm(self._data, points)  # type: ignore[reportGeneralTypeIssues]
        else:
            distance = cdist_minkowski(self._data, points, p=self.parameters.p)  # type: ignore[reportGeneralTypeIssues]

        return min_k(distance, k=n_neighbors)



from safecheck import Float64, Int64, NumpyArray, Real, typecheck
from sklearn.neighbors import NearestNeighbors as SklearnNearestNeighbors
from typing_extensions import Any, Literal

from ._base import NearestNeighbors


class SklearnNeighbors(NearestNeighbors):
    """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""

    available_algorithms = Literal["auto", "brute", "ball_tree", "kd_tree"]

    @typecheck
    def __init__(
        self,
        *,
        algorithm: available_algorithms = "auto",
        leaf_size: int = 30,
        metric: str = "minkowski",  # dynamically checked
        p: int = 2,
        metric_params: dict[str, Any] | None = None,
        n_jobs: int | None = None,
    ) -> None:
        super().__init__()
        # to be set in ``fit``
        self._model: None | SklearnNearestNeighbors = None

    @typecheck
    def fit(self, data: Real[NumpyArray, "n d"]) -> "SklearnNeighbors":
        self._model = SklearnNearestNeighbors(
            algorithm=self.parameters.algorithm,
            leaf_size=self.parameters.leaf_size,
            metric=self.parameters.metric,
            p=self.parameters.p,
            metric_params=self.parameters.metric_params,
            n_jobs=self.parameters.n_jobs,
        )
        self._model.fit(data)
        return self

    def query(
        self,
        point: Real[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float64[NumpyArray, "{n_neighbors}"]]:
        """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors=n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Real[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], Float64[NumpyArray, "m {n_neighbors}"]]:
        dist, idx = self._model.kneighbors(points, n_neighbors=n_neighbors)
        return idx, dist



# type: ignore[reportGeneralTypeIssues,reportIncompatibleMethodOverride]
# It appears to be impossible to correctly type the overrides without bloat.


import torch
from safecheck import AbstractDtype, Int64, NumpyArray, TorchArray, typecheck
from typing_extensions import Literal, overload

from ._base import NearestNeighbors


class SupportedFloat(AbstractDtype):
    dtypes = ["float32", "float64"]  # noqa: RUF012


class TorchNeighbors(NearestNeighbors):
    available_metrics = Literal["minkowski"]
    compute_modes = Literal[
        "use_mm_for_euclid_dist",
        "donot_use_mm_for_euclid_dist",
        "use_mm_for_euclid_dist_if_necessary",
    ]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "minkowski",
        p: int = 2,
        compute_mode: compute_modes = "use_mm_for_euclid_dist",
        force_dtype: torch.dtype | None = None,
        force_device: torch.device | str | None = None,
    ) -> None:
        super().__init__()
        # to be defined in ``fit``
        self._data: SupportedFloat[TorchArray, "n d"] | None = None

    @typecheck
    def fit(self, data: SupportedFloat[NumpyArray | TorchArray, "n d"]) -> "TorchNeighbors":
        self._data = torch.as_tensor(data, device=self.parameters.force_device, dtype=self.parameters.force_dtype)
        return self

    @overload
    def query(
        self,
        point: SupportedFloat[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], SupportedFloat[NumpyArray, "{n_neighbors}"]]: ...

    @overload
    def query(
        self,
        point: SupportedFloat[TorchArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[TorchArray, "{n_neighbors}"], SupportedFloat[TorchArray, "{n_neighbors}"]]: ...

    def query(self, point, n_neighbors):
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    @overload
    def query_batch(
        self,
        points: SupportedFloat[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], SupportedFloat[NumpyArray, "m {n_neighbors}"]]: ...

    @overload
    def query_batch(
        self,
        points: SupportedFloat[TorchArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[TorchArray, "m {n_neighbors}"], SupportedFloat[TorchArray, "m {n_neighbors}"]]: ...

    def query_batch(self, points, n_neighbors):
        is_numpy = isinstance(points, NumpyArray)
        points = torch.as_tensor(points, dtype=self._data.dtype, device=self._data.device)
        dist = torch.cdist(points, self._data, p=self.parameters.p, compute_mode=self.parameters.compute_mode)
        dist, idx = torch.topk(dist, k=n_neighbors, dim=1, largest=False)

        if is_numpy:
            return idx.cpu().numpy(), dist.cpu().numpy()

        return idx, dist



import logging

import faiss
from autofaiss import build_index
from safecheck import Float, Float32, Int64, NumpyArray, typecheck

from ._base import NearestNeighbors


class AutoFaissNeighbors(NearestNeighbors):
    """Similar to ``FaissNeighbors``, but the definition of the index is automated.

    Given the constraints provided in the arguments, ``autofaiss`` chooses an appropriate index.
    """

    @typecheck
    def __init__(
        self,
        *,
        save_on_disk: bool = False,
        pre_load_index: bool = False,
        pre_load_using_mmap: bool = False,
        index_path: str | None = None,
        index_infos_path: str | None = None,
        ids_path: str | None = None,
        file_format: str = "npy",
        embedding_column_name: str = "embedding",
        id_columns: list[str] | None = None,
        index_key: str | None = None,
        index_param: str | None = None,
        max_index_query_time_ms: int | float = 10.0,
        max_index_memory_usage: str = "16G",
        min_nearest_neighbors_to_retrieve: int = 20,
        current_memory_available: str = "32G",
        use_gpu: bool = False,
        metric_type: str = "ip",
        nb_cores: int | None = None,
        make_direct_map: bool = False,
        should_be_memory_mappable: bool = False,
        distributed: str | None = None,
        temporary_indices_folder: str = "hdfs://root/tmp/distributed_autofaiss_indices",
        verbose: int = logging.INFO,
        nb_indices_to_keep: int = 1,
    ) -> None:
        super().__init__()
        # to be defined in ``fit``
        self._index = None
        self._infos = None

        if pre_load_index:
            if pre_load_using_mmap:
                self._index = faiss.read_index(index_path, faiss.IO_FLAG_MMAP | faiss.IO_FLAG_READ_ONLY)
            else:
                self._index = faiss.read_index(index_path)
            self.is_fitted = True

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "AutoFaissNeighbors":
        self._index, self._infos = build_index(
            data,
            save_on_disk=self.parameters.save_on_disk,
            index_infos_path=self.parameters.index_infos_path,
            ids_path=self.parameters.ids_path,
            file_format=self.parameters.file_format,
            embedding_column_name=self.parameters.embedding_column_name,
            id_columns=self.parameters.id_columns,
            index_key=self.parameters.index_key,
            index_param=self.parameters.index_param,
            max_index_query_time_ms=self.parameters.max_index_query_time_ms,
            max_index_memory_usage=self.parameters.max_index_memory_usage,
            min_nearest_neighbors_to_retrieve=self.parameters.min_nearest_neighbors_to_retrieve,
            current_memory_available=self.parameters.current_memory_available,
            use_gpu=self.parameters.use_gpu,
            metric_type=self.parameters.metric_type,
            nb_cores=self.parameters.nb_cores,
            make_direct_map=self.parameters.make_direct_map,
            should_be_memory_mappable=self.parameters.should_be_memory_mappable,
            distributed=self.parameters.distributed,
            temporary_indices_folder=self.parameters.temporary_indices_folder,
            verbose=self.parameters.verbose,
            nb_indices_to_keep=self.parameters.nb_indices_to_keep,
        )
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float32[NumpyArray, "{n_neighbors}"]]:
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], Float32[NumpyArray, "m {n_neighbors}"]]:
        dist, idx = self._index.search(points, n_neighbors)  # type: ignore[reportGeneralTypeIssues]
        return idx, dist



from safecheck import Float, Int64, NumpyArray, typecheck
from scipy.spatial.distance import cdist
from typing_extensions import Literal, TypedDict

from ._base import NearestNeighbors
from ._numpy import min_k


class ScipyNeighbors(NearestNeighbors):
    """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""

    available_metrics = Literal[
        "braycurtis",
        "canberra",
        "chebyshev",
        "cityblock",
        "correlation",
        "cosine",
        "dice",
        "euclidean",
        "hamming",
        "jaccard",
        "jensenshannon",
        "kulczynski1",
        "mahalanobis",
        "matching",
        "minkowski",
        "rogerstanimoto",
        "russellrao",
        "seuclidean",
        "sokalmichener",
        "sokalsneath",
        "sqeuclidean",
        "yule",
    ]

    class MetricArgs(TypedDict):
        p: float
        w: NumpyArray
        V: NumpyArray
        VI: NumpyArray
        out: NumpyArray

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "euclidean",
        metric_args: MetricArgs | None = None,
    ) -> None:
        super().__init__()
        if metric_args is None:
            self.parameters.metric_args = {}

        # to be defined in ``fit``
        self._data: Float[NumpyArray, "n d"] | None = None

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "ScipyNeighbors":
        self._data = data
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float[NumpyArray, "{n_neighbors}"]]:
        """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], Float[NumpyArray, "m {n_neighbors}"]]:
        distance = cdist(
            points,
            self._data,  # type: ignore[reportGeneralTypeIssues]
            metric=self.parameters.metric,
            **self.parameters.metric_args,
        )
        idx, dist = min_k(distance, k=n_neighbors)
        return idx, dist



"""Entrypoint to the library exposing all available algorithms (if they are importable).

An algorithm can only be imported if the required libraries are installed. For example,
importing ``AnnoyNeighbors`` requires the ``annoy`` package to be available.

The additional libraries can be installed using nearness' package extras. For example,
to install nearness with ``annoy``, run ``pip install nearness[annoy]`` or install ``annoy``
by other means, for example, using ``conda install conda-forge::python-annoy``.

Installing nearness without any extras exposes only the abstract base class ``NearestNeighbors``
to define your own nearest neighbors algorithm. However, if you already have installed a
package that would otherwise be included as an extra dependency, you can use the corresponding
algorithm. For example, if you have installed ``nearness`` and ``numpy``, the algorithm
``NumpyNeighbors`` is exposed.
"""

from ._base import ExperimentalWarning, NearestNeighbors, config

__all__ = [
    "NearestNeighbors",
    "config",
    "ExperimentalWarning",
]


def get_version() -> str:
    """Return the package version or "unknown" if no version can be found."""
    from importlib import metadata

    try:
        return metadata.version(__name__)
    except metadata.PackageNotFoundError:  # pragma: no cover
        return "no-version-found-in-package-metadata"


try:
    from ._annoy import AnnoyNeighbors

    __all__ += ["AnnoyNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._autofaiss import AutoFaissNeighbors

    __all__ += ["AutoFaissNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._faiss import FaissNeighbors

    __all__ += ["FaissNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._hnsw import HNSWNeighbors

    __all__ += ["HNSWNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._jax import JaxNeighbors

    __all__ += ["JaxNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._numpy import NumpyNeighbors

    __all__ += ["NumpyNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._scann import ScannNeighbors

    __all__ += ["ScannNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._scipy import ScipyNeighbors

    __all__ += ["ScipyNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._sklearn import SklearnNeighbors

    __all__ += ["SklearnNeighbors"]
except ImportError:  # pragma: no cover
    ...

try:
    from ._torch import TorchNeighbors

    __all__ += ["TorchNeighbors"]
except ImportError:  # pragma: no cover
    ...



import os
import tempfile
from pathlib import Path
from uuid import uuid4

import numpy as np
from annoy import AnnoyIndex
from safecheck import Float, Int64, NumpyArray, Real, typecheck
from typing_extensions import Any, Iterable, Literal, overload

from ._base import NearestNeighbors

IS_WINDOWS = os.name == "nt"

__all__ = [
    "AnnoyNeighbors",
]


class AnnoyNeighbors(NearestNeighbors):
    available_metrics = Literal["angular", "euclidean", "manhattan", "hamming", "dot"]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "euclidean",
        n_trees: int = 32,
        n_search_neighbors: int = 128,
        random_seed: int | None = None,
        disk_build_path: str | Path | None = None,
        save_index_path: str | Path | None = None,
        load_index_path: str | Path | None = None,
        load_index_dim: int | None = None,
        prefault: bool = False,
    ) -> None:
        super().__init__()
        self._index: AnnoyIndex | None = None

        # mmap the raw annoy index
        if (path := self.parameters.load_index_path) is not None:
            index = AnnoyIndex(self.parameters.load_index_dim, self.parameters.metric)
            index.load(str(path))
            self._index = index
            self.is_fitted = True

    @overload
    def fit(self, data: Iterable[Real[NumpyArray, "d"]]) -> "AnnoyNeighbors": ...

    @overload
    def fit(self, data: Real[NumpyArray, "n d"]) -> "AnnoyNeighbors": ...

    def fit(self, data):  # type: ignore[reportGeneralTypeIssues]
        n_samples, n_dim = data.shape
        self.parameters.load_index_dim = n_dim
        index = AnnoyIndex(n_dim, metric=self.parameters.metric)

        if (seed := self.parameters.random_seed) is not None:
            index.set_seed(seed)

        if (path := self.parameters.disk_build_path) is not None:
            index.on_disk_build(str(path))

        for i in range(n_samples):
            index.add_item(i, data[i].tolist())

        index.build(self.parameters.n_trees)

        if (path := self.parameters.save_index_path) is not None:
            index.save(str(path), prefault=self.parameters.prefault)

        self._index = index
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float[NumpyArray, "{n_neighbors}"]]:
        """No runtime type checks, because query is used in the ``query_batch`` loop."""
        idx, dist = self._index.get_nns_by_vector(
            point.tolist(),
            n_neighbors,
            self.parameters.n_search_neighbors,
            include_distances=True,
        )
        return np.array(idx, dtype=np.int64), np.array(dist, dtype=point.dtype)

    def query_idx(self, point: Float[NumpyArray, "d"], n_neighbors: int) -> Int64[NumpyArray, "{n_neighbors}"]:
        idx = self._index.get_nns_by_vector(
            point.tolist(),
            n_neighbors,
            self.parameters.n_search_neighbors,
        )
        return np.array(idx, dtype=np.int64)

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Load the index using a temporary file."""
        super().__setstate__(state)
        if self._index is not None:
            index_bytes = state.pop("_index")
            self._index = AnnoyIndex(self.parameters.load_index_dim, self.parameters.metric)
            # cannot delete the file on windows, ``delete_on_close`` is only available with Python >= 3.12
            with tempfile.NamedTemporaryFile("wb", delete=not IS_WINDOWS) as file:
                file.write(index_bytes)
                file.flush()  # otherwise errors with 'Inappropriate ioctl for device'

                # it's not allowed to re-open the file on windows
                # https://stackoverflow.com/questions/23212435/permission-denied-to-write-to-my-temporary-file
                if IS_WINDOWS:
                    file.close()

                # open the file and load the index
                self._index.load(file.name, prefault=self.parameters.prefault)

    def __getstate__(self) -> dict[str, Any]:
        """Save the index to a temporary file and store the bytes of the file."""
        state = super().__getstate__()
        if self._index is not None:
            # there might be cleanup errors in multiprocessing setting on windows
            with tempfile.TemporaryDirectory(ignore_cleanup_errors=IS_WINDOWS) as tmp_dir:
                file_path = Path(tmp_dir) / uuid4().hex
                self._index.save(str(file_path), prefault=self.parameters.prefault)
                state["_index"] = file_path.read_bytes()
        return state



import warnings
from functools import wraps

__all__ = [
    "experimental",
    "ExperimentalWarning",
]

from typing_extensions import Any, Callable

EXPERIMENTAL_WARNING_HELP = (
    "To mute warnings for experimental functionality, invoke"
    ' warnings.filterwarnings("ignore", category=nearness.ExperimentalWarning) or use'
    " one of the other methods described at"
    " https://docs.python.org/3/library/warnings.html#describing-warning-filters."
)


def experimental(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        experimental_warning(func.__name__)
        return func(*args, **kwargs)

    return wrapper


def experimental_warning(
    subject: str,
    additional_warn_text: str | None = None,
    stacklevel: int = 3,
) -> None:
    extra_text = f" {additional_warn_text}" if additional_warn_text else ""
    warnings.warn(
        (
            f"{subject} is experimental. It may break in future versions, even between dot"
            f" releases.{extra_text} {EXPERIMENTAL_WARNING_HELP}"
        ),
        ExperimentalWarning,
        stacklevel=stacklevel,
    )


class ExperimentalWarning(Warning):
    pass



from hnswlib import BFIndex, Index
from safecheck import Float, Float32, NumpyArray, UInt64, typecheck
from typing_extensions import Literal

from ._base import NearestNeighbors


class HNSWNeighbors(NearestNeighbors):
    """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""

    available_metrics = Literal["l2", "ip", "cosine"]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "l2",
        n_index_neighbors: int = 256,
        n_search_neighbors: int | None = None,
        n_links: int = 16,
        n_threads: int = -1,
        random_seed: int = 0,
        use_bruteforce: bool = False,
        load_dim: int | None = None,
    ) -> None:
        super().__init__()
        self._index_constructor = BFIndex if use_bruteforce else Index
        self._model = None

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "HNSWNeighbors":
        n_samples, n_dim = data.shape
        index = self._index_constructor(space=self.parameters.metric, dim=n_dim)

        if self.parameters.use_bruteforce:
            index.init_index(max_elements=n_samples)
            index.add_items(data)
        else:
            index.init_index(
                max_elements=n_samples,
                ef_construction=self.parameters.n_index_neighbors,
                M=self.parameters.n_links,
                random_seed=self.parameters.random_seed,
            )
            index.add_items(data)

        if (n_search := self.parameters.n_search_neighbors) is not None:
            index.set_ef(n_search)

        self._model = index
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[UInt64[NumpyArray, "{n_neighbors}"], Float32[NumpyArray, "{n_neighbors}"]]:
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[UInt64[NumpyArray, "m {n_neighbors}"], Float32[NumpyArray, "m {n_neighbors}"]]:
        idx, dist = self._model.knn_query(points, k=n_neighbors, num_threads=self.parameters.n_threads)
        return idx, dist



from dataclasses import asdict, dataclass

from typing_extensions import Any, Callable

__all__ = ["Config", "config"]


@dataclass
class Config:
    """The global NearestNeighbors configuration.

    Each instance of ``NearestNeighbors`` creates a current (deep) copy of the global ``config`` object. It is,
    therefore, possible to adapt class behaviour before instantiation by modification of ``config``, and, because
    each class owns a copy of ``config``, it can be configured individually.

    We further provide a simple event interface using callbacks. You can observe ``__setattr__`` events using
    ``register_callback`` and provide a function receiving the newly set value. To enable the callback-based
    observability it's necessary that all fields are immutable, the only way to update a field is to set it.
    """

    def __init__(self) -> None:
        super().__init__()
        self._callbacks = {}

    methods_require_fit: frozenset[str] = frozenset(
        {
            "query",
            "query_idx",
            "query_dist",
            "query_batch",
            "query_batch_idx",
            "query_batch_dist",
        },
    )
    save_protocol: int | None = None
    save_compression: int = 0

    def register_callback(self, field_name: str, callback: Callable[[Any], Any]) -> None:
        if field_name in asdict(self):
            # initialize the list of callbacks if there is no callback yet
            if field_name not in self._callbacks:
                self._callbacks[field_name] = []
            self._callbacks[field_name].append(callback)
        else:
            msg = (
                f"Cannot register a callback for attribute '{field_name}', attribute "
                f"does not exist on class {self}."
            )
            raise AssertionError(msg)

    def __setattr__(self, name: str, value: Any) -> None:
        super().__setattr__(name, value)
        if name in self.__annotations__:
            callbacks = self._callbacks.get(name, [])
            for callback in callbacks:
                callback(value)


config = Config()



import inspect
import threading
import types
from abc import ABCMeta, abstractmethod
from copy import deepcopy
from dataclasses import fields, make_dataclass
from functools import partial, wraps
from logging import getLogger
from pathlib import Path
from warnings import warn

import joblib
import numpy as np
from joblib import Parallel, delayed
from safecheck import typecheck
from typing_extensions import Any, Callable, Self

from ._config import Config, config
from ._experimental import experimental

logger = getLogger(__name__)

__all__ = [
    "NearestNeighbors",
    "InvalidSignatureError",
]


class InvalidSignatureError(ValueError): ...


class NearestNeighborsMeta(ABCMeta):
    """Metaclass for NearestNeighbors.

    This metaclass is used to:
    1. Check if the implemented ``__init__`` only uses keyword-only parameters.
    2. Dynamically set an attribute ``_parameters_`` containing all ``__init__`` parameters.
    3. Dynamically set an attribute ``_config_`` containing a copy of the global ``config`` object.

    The parameters are modeled as a ``dataclasses.dataclass``, therefore, allowing:
    1. Parameter access using ``instance.parameters.name`` or ``getattr(instance.parameters, name)``
    2. Parameter updates using ``instance.parameters.name = value`` or ``setattr(instance.parameters, name, value)``
    """

    def __new__(cls, name: str, bases: tuple[type, ...], attrs: dict[str, Any]) -> type:
        """Check the signature of ``__init__`` to ensure keyword-only arguments."""
        if "__init__" in attrs:
            parameters = inspect.signature(attrs["__init__"]).parameters
            for name, parameter in parameters.items():
                if name != "self" and (kind := parameter.kind) is not inspect.Parameter.KEYWORD_ONLY:
                    msg = (
                        "Only keyword-only arguments are allowed for classes inheriting from 'nearness."
                        f"NearestNeighbors', but found parameter '{parameter}' of kind '{kind}'."
                        f" Hint: You can enforce keyword-only arguments using the 'single star' syntax before the"
                        f" parameters, e.g. '___init___(self, *, {parameter}, ...), see "
                        f"https://peps.python.org/pep-3102/."
                    )
                    raise InvalidSignatureError(msg)

        return super().__new__(cls, name, bases, attrs)

    def __call__(cls, *_: Any, **kwargs: Any) -> "NearestNeighbors":
        """Dynamically set the ``_parameters_`` attribute and wrap the fit method and query methods.

        :param _: We ensure that there are only keyword-only arguments in ``__new__``.
        :param kwargs: The mapping of parameters of the algorithm.
        :return: The modified object.
        """
        # filter the self parameter
        parameters = {k: v for k, v in inspect.signature(cls.__init__).parameters.items() if k != "self"}

        # We don't want threads to override class-bound ``_parameters_`` or ``_config_``.
        # Because we use thread local storage, the ``_config_`` and ``_parameters`` attribute on the class should
        # never be visible to another parallel class instantiation.
        thread_local = threading.local()

        # Set the default config, such that it can be manipulated in ``__init__``, and deepcopy the config such that
        # a manipulation of ``cls._config_`` does not manipulate the global ``config`` object.
        # We must set the attribute here before ``__call__``, because the config should be usable in ``__init__``.
        thread_local.config = deepcopy(config)
        cls._config_ = thread_local.config  # type: ignore[reportUninitializedInstanceVariable]
        logger.info("Determined config object as '%s'", cls._config_)

        # same for the parameters, we set the parameters before the call such that they are usable in ``__init__``
        thread_local.parameters = _create_parameter_class(parameters, kwargs)
        cls._parameters_ = thread_local.parameters  # type: ignore[reportUninitializedInstanceVariable]
        logger.info("Determined parameters as '%s'", cls._parameters_)

        # now we set all the relevant attributes on the ``instance``, as they should not be class-bound.
        # order is important here as ``__wrap_fit_method`` and ``__wrap_check_method`` depend on the set attributes
        obj = type.__call__(cls, **kwargs)
        obj._parameters_, obj._config_ = cls._parameters_, cls._config_  # noqa: SLF001
        # make sure that the wrapped methods are in sync when the config is changed after class instantiation
        obj._config_.register_callback(  # noqa: SLF001
            "methods_require_fit",
            partial(cls._check_callback, obj=obj),
        )
        cls._wrap_fit_method(obj)
        cls._wrap_check_method(obj)
        del cls._parameters_
        del cls._config_
        return obj

    def _check_callback(cls, _: Any, *, obj: "NearestNeighbors") -> None:
        """A callback to refresh the ``__fitted__`` attributes when the 'methods_require_fit' attribute is set ."""
        if not obj.is_fitted:  # if the object is already fitted there is no need to wrap any methods
            cls._wrap_check_method(obj)

    def _wrap_fit_method(cls, obj: "NearestNeighbors") -> None:
        """Wrap the ``fit`` method to ensure it sets the ``__fitted__`` attribute and unwraps the query methods.

        This feels a bit like magic, but the alternatives would be to:
        1. Add a decorator to every ``fit`` method that sets ``__fitted__``.
        2. Set the ``__fitted__`` attribute on every ``fit``-like method and check for ``__fitted__`` everywhere.
        """
        fit = obj.fit

        @wraps(fit)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            result = fit(*args, **kwargs)
            logger.debug("Called wrapped fit method with successful fit result, setting '__fitted__'.")
            obj.is_fitted = True
            return result

        # we set a ``__fit__`` attribute to denote that the method has been decorated
        wrapper.__fit__ = True  # type: ignore[reportGeneralTypeIssues]
        obj.fit = wrapper  # override the default fit with the wrapped fit

    def _wrap_check_method(cls, obj: "NearestNeighbors") -> None:
        """Wrap or unwrap methods according to ``methods_require_fit``.

        For each wrapped method, we ensure that ``__fitted__`` is True, before the method can be called.
        Methods not in ``methods_require_fit`` are unwrapped.
        """
        logger.debug("Starting to wrap methods to enable fit checking.")
        available_attributes = dir(obj)
        methods_to_wrap = obj._config_.methods_require_fit  # noqa: SLF001
        for attribute_name in available_attributes:
            attribute = getattr(obj, attribute_name)
            has_check = hasattr(attribute, "__check__")
            if attribute_name in methods_to_wrap:
                if has_check:
                    continue
                # check if the ``method`` attribute is a bound method
                if isinstance(attribute, types.MethodType):
                    logger.debug("Wrapping attribute '%s'.", attribute_name)
                    # update the method with the wrapped ``__fitted__`` check
                    setattr(obj, attribute_name, wraps(attribute)(_create_check_wrapper(obj, attribute)))
                else:
                    msg = (
                        f"Attempting to enable '__fitted__' check for invalid attribute '{attribute_name}', because "
                        f"'{attribute_name}' is not a bound method, instead it's an attribute of '{type(attribute)}'."
                    )
                    warn(msg, stacklevel=1)
            elif has_check:  # if an attribute has a ``__check__`` attribute, but is not in the set to check, we unwrap.
                logger.debug(
                    "Attribute '%s' is not in '%s', unwrapping the attribute.",
                    attribute_name,
                    methods_to_wrap,
                )
                setattr(obj, attribute_name, attribute.__wrapped__)

        for attribute_name in methods_to_wrap:
            if attribute_name not in available_attributes:
                msg = (
                    f"Attempting to enable '__fitted__' check for missing attribute '{attribute_name}', because "
                    f"'{attribute_name}' does not exist on class {obj}."
                )
                warn(msg, stacklevel=1)

    def _unwrap_check_method(cls, obj: "NearestNeighbors") -> None:
        """Unwrap all existing methods in ``methods_require_fit`` to disable the ``__fitted__`` check.

        This is just a performance optimization, it retrieves the original method and removes the implicitly
        generated function wrapper (decorator). It should be safe to unwrap the methods if ``__fitted__``
        is only set in ``fit``, but unsafe when ``__fitted__`` is manually set to ``False`` after ``fit``.
        """
        logger.debug("Starting to unwrap all fit checking methods.")
        for method_name in obj._config_.methods_require_fit:  # noqa: SLF001
            # we set an __requires_fit__ attribute on the wrapper, because using ``__wrapped__`` alone is not
            # safe (methods also use ``__wrapped__`` starting with Python 3.10)
            if hasattr(obj, method_name) and hasattr(method := getattr(obj, method_name), "__check__"):
                logger.debug("Unwrapping method '%s'.", method_name)
                setattr(obj, method_name, method.__wrapped__)


class NearestNeighbors(metaclass=NearestNeighborsMeta):
    """Abstract base class for nearest neighbors search algorithms.

    The minimal implementation consists of a ``fit`` and ``query`` method. The default ``query_batch`` method uses
    a thread pool to parallelize calls to ``query``, but all methods may be overridden for efficiency purposes.
    It is expected that ``query`` returns a tuple of indices and distances to the nearest neighbors.

    The default interface to all methods is based on NumPy N-dimensional arrays, but implementations might ``overload``
    the methods such that other data types can be implemented. An important consideration for all possible methods
    and overloads is that they should be type stable, preferably allowing only floating-point arrays as input
    and returning floating-point distances of equal type as output.
    """

    __fitted__ = False

    @abstractmethod
    def fit(self, data: np.ndarray) -> Self:
        """Learn an index structure based on a matrix of points.

        :param data: matrix of ``size x dim``.
        :return: reference to object (``self``).
        """
        ...

    @abstractmethod
    def query(self, point: np.ndarray, n_neighbors: int) -> tuple[np.ndarray, np.ndarray]:
        """Search ``n_neighbors`` for a single point, returning the indices and distances.

        :param point: vector of ``dim``.
        :param n_neighbors: number of neighbors to search.
        :return:  (vector of indices, vector of distances) of size ``n_neighbors``
        """
        ...

    @experimental
    def query_idx(self, point: np.ndarray, n_neighbors: int) -> np.ndarray:
        """Search ``n_neighbors`` for a single point, returning the indices.

        :param point: vector of ``dim``.
        :param n_neighbors: number of neighbors to search.
        :return: vector of indices of size ``n_neighbors``
        """
        idx, _ = self.query(point, n_neighbors)
        return idx

    @experimental
    def query_dist(self, point: np.ndarray, n_neighbors: int) -> np.ndarray:
        """Search ``n_neighbors`` for a single point (vector), returning the distances.

        :param point: vector of ``dim``.
        :param n_neighbors: number of neighbors to search.
        :return: vector of distances of size ``n_neighbors``
        """
        _, dist = self.query(point, n_neighbors)
        return dist

    def query_batch(self, points: np.ndarray, n_neighbors: int) -> tuple[np.ndarray, np.ndarray]:
        """Search ``n_neighbors`` for a batch of points, returning the indices and distances.

        :param points: matrix of ``batch x dim``.
        :param n_neighbors: number of neighbors to search.
        :return: (matrix of indices, matrix of distances) of size ``batch x n_neighbors``.
        """
        result = Parallel(prefer="threads")(delayed(self.query)(q, n_neighbors) for q in points)
        idx, dist = zip(*result)
        return np.stack(idx), np.stack(dist)

    @experimental
    def query_batch_idx(self, points: np.ndarray, n_neighbors: int) -> np.ndarray:
        """Search ``n_neighbors`` for a batch of points, returning the indices.

        :param points: matrix of ``batch x dim``.
        :param n_neighbors: number of neighbors to search.
        :return: matrix of indices ``batch x n_neighbors``.
        """
        idx, _ = self.query_batch(points, n_neighbors)
        return idx

    @experimental
    def query_batch_dist(self, points: np.ndarray, n_neighbors: int) -> np.ndarray:
        """Search ``n_neighbors`` for a batch of points, returning the distances.

        :param points: matrix of ``batch x dim``.
        :param n_neighbors: number of neighbors to search.
        :return: matrix of distances ``batch x n_neighbors``.
        """
        _, dist = self.query_batch(points, n_neighbors)
        return dist

    def add(self, data: np.ndarray) -> Self:
        """Partially add elements to the index.

        Not all methods implement ``add``, but this can be useful if the data does not fit into memory.

        :param data: Data to add to the index.
        :raises: NotImplementedError if ``add`` is not available for the class.
        """
        msg = (
            f"Method 'add' is not implemented for '{self}', use 'fit' to learn the entire database at once, "
            f"or use a different indexing structure that enables partial addition of data. "
            f"If you would like to implement 'add' for '{self}', please open a pull request at "
            f"https://github.com/davnn/nearness"
        )
        raise NotImplementedError(msg)

    def save(self, file: str | Path) -> None:
        """Save the state of the model using pickle such that it can be fully restored using ``load``.

        :param file: name or path of the file to save to.
        :return: nothing.
        """
        joblib.dump(
            self,
            filename=file,
            protocol=self.config.save_protocol,
            compress=self.config.save_compression,
        )

    @staticmethod
    def load(file: str | Path) -> "NearestNeighbors":
        """Load a model using pickle to fully restore the saved state.

        :param file: name or path of the file to load from.
        :return: the restored ``NearestNeighbors`` algorithm.
        """
        return joblib.load(file)

    def __setstate__(self, state: dict[str, Any]) -> None:
        parameter_types, parameter_values = state["_parameters_"]
        state["_parameters_"] = make_dataclass("Parameters", parameter_types)(**parameter_values)
        self.__dict__.update(state)
        NearestNeighbors._wrap_fit_method(self)  # re-wrap the fit method
        self.is_fitted = self.__fitted__  # wrap or unwrap the check methods depending on ``__fitted__``

    def __getstate__(self) -> dict[str, Any]:
        self.fit = (  # unwrap the fit method if wrapped
            self.fit.__wrapped__  # type: ignore[reportFunctionMemberAccess]
            if hasattr(self.fit, "__fit__")
            else self.fit
        )
        NearestNeighbors._unwrap_check_method(self)
        state = self.__dict__.copy()
        parameter_fields = fields(self._parameters_)
        parameter_types = [(f.name, f.type) for f in parameter_fields]
        parameter_values = {f.name: getattr(self._parameters_, f.name) for f in parameter_fields}
        state["_parameters_"] = (parameter_types, parameter_values)
        return state

    @property
    def is_fitted(self) -> bool:
        return self.__fitted__

    @is_fitted.setter
    @typecheck
    def is_fitted(self, value: bool) -> None:
        if value:
            NearestNeighbors._unwrap_check_method(self)

        if not value:
            NearestNeighbors._wrap_check_method(self)

        self.__fitted__ = value

    @property
    def config(self) -> "Config":
        return self._config_

    @property
    def parameters(self) -> Any:
        """The parameters are dynamically set on class creation due to the ``NearestNeighborsMeta`` metaclass.

        All ``__init__`` arguments are considered parameters, and it is suggested to type-hint every parameter.
        """
        return self._parameters_


def _create_parameter_class(
    parameters: dict[str, inspect.Parameter],
    kwargs: dict[str, Any],
) -> "types.Parameters":  # type: ignore[reportGeneralTypeIssues]
    empty = inspect.Parameter.empty
    parameter_types = [(k, Any if (a := v.annotation) is empty else a) for k, v in parameters.items()]
    parameter_values = {k: kwargs.get(k, v.default) for k, v in parameters.items()}
    return make_dataclass("Parameters", parameter_types)(**parameter_values)  # type: ignore[reportArgumentType]


def _create_check_wrapper(obj: NearestNeighbors, method: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(method)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        logger.debug("Called method with fit check enabled.")
        if not obj.is_fitted:
            msg = f"Attempted to call '{method.__qualname__}', but 'fit' has not yet been called."
            raise AssertionError(msg)
        return method(*args, **kwargs)

    # we set a ``__check__`` attribute, to be able to (easily) distinguish the wrapper from the bound method
    wrapper.__check__ = True  # type: ignore[reportFunctionMemberAccess]
    return wrapper



import faiss
import numpy as np
from safecheck import Float, Float32, Int64, NumpyArray, typecheck
from typing_extensions import Any, Protocol, runtime_checkable

from ._base import NearestNeighbors


@runtime_checkable
class FaissIndexFactory(Protocol):
    def __call__(self, dim: int) -> faiss.Index: ...


class FaissNeighbors(NearestNeighbors):
    """A fairly simple wrapper around ``faiss.Index``.

    A Faiss index is built from any combination of the following.

    1. Vector transform — a pre-processing step applied to vectors before indexing (PCA, OPQ).
    2. Coarse quantizer — rough organization of vectors to subdomains (for restricting search
       scope, includes IVF, IMI, and HNSW).
    3. Fine quantizer — a finer compression of vectors into smaller domains (for compressing
       index size, such as PQ).
    4. Refinement — a final step at search-time which re-orders results using distance calculations
       on the original flat vectors. Alternatively, another index (non-flat) index can be used.

    Note that coarse quantization refers to the 'clustering' of vectors (such as inverted indexing with IVF).
    By using coarse quantization, we enable non-exhaustive search by limiting the search scope.
    Fine quantization describes the compression of vectors into codes (as with PQ). The purpose fine quantization
    is to reduce the memory usage of the index.

    References
    ----------
        - https://github.com/facebookresearch/faiss
        - https://www.pinecone.io/learn/series/faiss/

    """

    @typecheck
    def __init__(
        self,
        *,
        index_or_factory: str | FaissIndexFactory | faiss.Index = "Flat",
        add_data_on_fit: bool = True,
        sample_train_points: int | float | None = None,
        sample_with_replacement: bool = False,
        rng: Any = None,  # types are validated by default_rng
    ) -> None:
        super().__init__()
        self.parameters.rng = rng if isinstance(rng, np.random.Generator) else np.random.default_rng(rng)

        # to be defined in ``fit``
        self._index: faiss.Index | None = None

    @typecheck
    def fit(self, data: Float[NumpyArray, "n d"]) -> "FaissNeighbors":
        n_samples, dim = data.shape
        self._index = self._create_index(dim)
        if (n_train := self.parameters.sample_train_points) is not None:
            if isinstance(n_train, float):
                # we make sure float is 0 < float < 1 in ``__init__``
                n_train = int(n_train * n_samples)

            sample = self.parameters.rng.choice(data, size=n_train, replace=self.parameters.sample_with_replacement)
            self._index.train(sample)  # type: ignore[reportGeneralTypeIssues]
        else:
            self._index.train(data)  # type: ignore[reportGeneralTypeIssues]

        if self.parameters.add_data_on_fit:
            # data might be added directly on fit, or using the ``add`` method
            self._index.add(data)  # type: ignore[reportGeneralTypeIssues]

        return self

    @typecheck
    def add(self, data: Float[NumpyArray, "n d"]) -> "FaissNeighbors":
        self._index.add(data)  # type: ignore[reportGeneralTypeIssues]
        return self

    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "{n_neighbors}"], Float32[NumpyArray, "{n_neighbors}"]]:
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int64[NumpyArray, "m {n_neighbors}"], Float32[NumpyArray, "m {n_neighbors}"]]:
        dist, idx = self._index.search(points, n_neighbors)  # type: ignore[reportGeneralTypeIssues]
        return idx, dist

    def _create_index(self, dim: int) -> faiss.Index:
        index = self.parameters.index_or_factory
        if isinstance(index, str):
            # if the index is a string, we treat it as an index factory
            return faiss.index_factory(dim, index)

        if isinstance(index, faiss.Index):
            # if the index is an existing index, simply use that
            return index

        # otherwise the index should be a callable returning an index
        return index(dim)



import jax
import jax.numpy as jnp
import numpy as np
from safecheck import Float, Float32, Int32, JaxArray, NumpyArray, typecheck
from typing_extensions import Literal, overload

from ._base import NearestNeighbors


def cdist_minkowski(
    data: Float[NumpyArray | JaxArray, "n d"],
    query: Float[NumpyArray | JaxArray, "m d"],
    p: int | float = 2,
) -> Float[NumpyArray | JaxArray, "m n"]:
    """Reference implementation of Euclidean distance with Jax.

    :param data: Matrix of input data points.
    :param query: Matrix of query data points.
    :param p: The power parameter for the Minkowski distance metric.
    :return: Matrix of distances between query and data.
    """
    diff = jnp.expand_dims(query, 1) - jnp.expand_dims(data, 0)
    return jnp.linalg.norm(diff, ord=p, axis=-1, keepdims=False)


def cdist_euclidean_mm(
    data: Float[NumpyArray | JaxArray, "n d"],
    query: Float[NumpyArray | JaxArray, "m d"],
) -> Float[NumpyArray | JaxArray, "m n"]:
    """Probably the fastest possible implementation of Euclidean distance with NumPy.

    :param data: Matrix of input data points.
    :param query: Matrix of query data points.
    :return: Matrix of distances between query and data.
    """
    # Calculate the squared differences raised to the power of p
    squared_diff_p = (
        jnp.expand_dims(jnp.einsum("ij,ij->i", data, data), 0)
        - 2.0 * jnp.dot(query, data.T)
        + jnp.expand_dims(jnp.einsum("ij,ij->i", query, query), 1)
    )

    # Ensure that negative values close to zero are treated as zeros due to potential numerical errors
    return jnp.sqrt(jnp.maximum(squared_diff_p, 0.0))


class JaxNeighbors(NearestNeighbors):
    """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""

    available_metrics = Literal["minkowski"]
    compute_modes = Literal["use_mm_for_euclid_dist", "donot_use_mm_for_euclid_dist"]

    @typecheck
    def __init__(
        self,
        *,
        metric: available_metrics = "minkowski",
        p: int = 2,
        compute_mode: compute_modes = "use_mm_for_euclid_dist",
        approximate_recall_target: float = 0.95,
    ) -> None:
        super().__init__()
        # to be defined in ``fit``
        self._data: Float[JaxArray, "n d"] | None = None

    @typecheck
    def fit(self, data: Float[NumpyArray | JaxArray, "n d"]) -> "JaxNeighbors":
        self._data = jnp.asarray(data)
        return self

    @overload
    def query(
        self,
        point: Float[JaxArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int32[JaxArray, "{n_neighbors}"], Float32[JaxArray, "{n_neighbors}"]]: ...

    @overload
    def query(
        self,
        point: Float[NumpyArray, "d"],
        n_neighbors: int,
    ) -> tuple[Int32[NumpyArray, "{n_neighbors}"], Float32[NumpyArray, "{n_neighbors}"]]: ...

    def query(self, point, n_neighbors):  # type: ignore[reportGeneralTypeIssues]
        """CPU-based nearest neighbors algorithm based on scikit-learn. Note: The distances and indices are sorted!."""
        idx, dist = self.query_batch(point.reshape(1, -1), n_neighbors)
        return idx.ravel(), dist.ravel()

    @overload
    def query_batch(
        self,
        points: Float[JaxArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int32[JaxArray, "m {n_neighbors}"], Float32[JaxArray, "m {n_neighbors}"]]: ...

    @overload
    def query_batch(
        self,
        points: Float[NumpyArray, "m d"],
        n_neighbors: int,
    ) -> tuple[Int32[NumpyArray, "m {n_neighbors}"], Float32[NumpyArray, "m {n_neighbors}"]]: ...

    def query_batch(self, points, n_neighbors):  # type: ignore[reportGeneralTypeIssues]
        # the data is implicitly casted to a jax array
        is_numpy = isinstance(points, NumpyArray)  # type: ignore[reportUnnecessaryIsInstance,reportGeneralTypeIssues]
        if self.parameters.p == 2 and self.parameters.compute_mode == "use_mm_for_euclid_dist":
            distance = cdist_euclidean_mm(self._data, points)  # type: ignore[reportGeneralTypeIssues]
        else:
            distance = cdist_minkowski(self._data, points, p=self.parameters.p)  # type: ignore[reportGeneralTypeIssues]

        dist, idx = jax.lax.approx_min_k(
            distance,  # type: ignore[reportGeneralTypeIssues]
            n_neighbors,
            recall_target=self.parameters.approximate_recall_target,
        )
        # ``np.asarray`` should be the best option to ensure no copy is made, https://github.com/google/jax/issues/1961
        if is_numpy:
            return np.asarray(idx), np.asarray(dist)

        return idx, dist
